# Copyright (C) 2022-2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import datetime
import json
from pathlib import Path
from typing import Type

import pkg_resources

from swh.loader.core.metadata_fetchers import MetadataFetcherProtocol
from swh.loader.metadata import __version__
from swh.loader.metadata.gitea import GiteaMetadataFetcher
from swh.model.model import (
    MetadataAuthority,
    MetadataAuthorityType,
    MetadataFetcher,
    Origin,
    RawExtrinsicMetadata,
)

from .test_base import DummyLoader

ORIGIN = Origin("https://codeberg.org/ForgeFed/ForgeFed")
FORKED_ORIGIN = Origin("https://codeberg.org/_ZN3val/ForgeFed")
ORIGIN_WITH_PATH_PREFIX = Origin(
    "https://git.osgeo.org/gitea/postgis/postgis-workshops"
)

METADATA_AUTHORITY = MetadataAuthority(
    type=MetadataAuthorityType.FORGE, url="https://codeberg.org"
)


def expected_metadata(dt, datadir):
    data_file_path = Path(datadir) / "https_codeberg.org/api_v1_repos_ForgeFed_ForgeFed"
    with data_file_path.open("rb") as fd:
        expected_metadata_bytes = fd.read()
    return RawExtrinsicMetadata(
        target=ORIGIN.swhid(),
        discovery_date=dt,
        authority=METADATA_AUTHORITY,
        fetcher=MetadataFetcher(name="swh.loader.metadata.gitea", version=__version__),
        format="gitea-repository-json",
        metadata=expected_metadata_bytes,
    )


def test_type() -> None:
    # check with mypy
    fetcher_cls: Type[MetadataFetcherProtocol]
    fetcher_cls = GiteaMetadataFetcher
    print(fetcher_cls)

    # check at runtime
    fetcher = GiteaMetadataFetcher(
        ORIGIN,
        credentials=None,
        lister_name="gitea",
        lister_instance_name="gitea",
    )
    assert isinstance(fetcher, MetadataFetcherProtocol)


def test_gitea_metadata(datadir, requests_mock_datadir, mocker):
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    mocker.patch("swh.loader.metadata.base._now", return_value=now)

    fetcher = GiteaMetadataFetcher(
        ORIGIN, credentials=None, lister_name="gitea", lister_instance_name="gitea"
    )

    assert fetcher.get_origin_metadata()[0] == expected_metadata(now, datadir)
    assert fetcher.get_origin_metadata() == [expected_metadata(now, datadir)]
    assert fetcher.get_parent_origins() == []

    # Need to make sure '/' in the project ID was encoded as %2F; as
    # requests_mock_datadir does not tell the difference (because it uses urlunquote,
    # which decodes it)
    assert len(requests_mock_datadir.request_history) == 1
    assert (
        requests_mock_datadir.request_history[0].url
        == "https://codeberg.org/api/v1/repos/ForgeFed/ForgeFed"
    )


def test_gitea_metadata_fork(datadir, requests_mock_datadir, mocker):
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    mocker.patch("swh.loader.metadata.base._now", return_value=now)

    fetcher = GiteaMetadataFetcher(
        FORKED_ORIGIN,
        credentials=None,
        lister_name="gitea",
        lister_instance_name="gitea",
    )

    assert fetcher.get_parent_origins() == [ORIGIN]


def test_gitea_metadata_from_loader(
    swh_storage, mocker, datadir, requests_mock_datadir
):
    # Fail early if this package is not fully installed
    assert "gitea" in {
        entry_point.name
        for entry_point in pkg_resources.iter_entry_points("swh.loader.metadata")
    }

    now = datetime.datetime.now(tz=datetime.timezone.utc)
    mocker.patch("swh.loader.metadata.base._now", return_value=now)

    loader = DummyLoader(
        storage=swh_storage,
        origin_url=ORIGIN.url,
        lister_name="gitea",
        lister_instance_name="gitea",
    )
    loader.load()

    assert swh_storage.raw_extrinsic_metadata_get(
        ORIGIN.swhid(), METADATA_AUTHORITY
    ).results == [expected_metadata(now, datadir)]


def test_gitea_metadata_origin_with_path_prefix(datadir, requests_mock_datadir, mocker):
    now = datetime.datetime.now(tz=datetime.timezone.utc)
    mocker.patch("swh.loader.metadata.base._now", return_value=now)

    fetcher = GiteaMetadataFetcher(
        ORIGIN_WITH_PATH_PREFIX,
        credentials=None,
        lister_name="gitea",
        lister_instance_name="gitea",
    )

    metadata = fetcher.get_origin_metadata()

    assert (
        metadata
        and json.loads(metadata[0].metadata).get("html_url")
        == ORIGIN_WITH_PATH_PREFIX.url
    )



# Copyright (C) 2023  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

# WARNING: do not import unnecessary things here to keep cli startup time under
# control
import click

from swh.core.cli import CONTEXT_SETTINGS
from swh.core.cli import swh as swh_cli_group


@swh_cli_group.group(name="metadata-loader", context_settings=CONTEXT_SETTINGS)
@click.option(
    "--config-file",
    "-C",
    default=None,
    type=click.Path(
        exists=True,
        dir_okay=False,
    ),
    help="Configuration file.",
)
@click.pass_context
def metadata_loader_cli_group(ctx, config_file):
    """Software Heritage Metadata Loader.

    The Indexer is used to mine the content of the archive and extract derived
    information from archive source code artifacts.

    """
    from swh.core import config

    ctx.ensure_object(dict)
    conf = config.read(config_file)
    ctx.obj["config"] = conf


@metadata_loader_cli_group.command("journal-client")
@click.pass_context
@click.option(
    "--stop-after-objects",
    "-m",
    default=None,
    type=int,
    help="Maximum number of objects to replay. Default is to run forever.",
)
def journal_client(ctx, stop_after_objects: int) -> None:
    """Load metadata from origins not visited for a while.

    Required configuration keys:

    * ``scheduler``
    * ``journal``
    * ``storage``
    * ``reload_after_days``, the number of days such that, if there is already metadata
      less old than this, then metadata is not loaded again
    * ``metadata_fetcher_credentials``, in the same format as for loaders and listers,
      possibly empty
    """
    from swh.journal.client import get_journal_client
    from swh.loader.metadata.journal_client import JournalClient
    from swh.scheduler import get_scheduler
    from swh.storage import get_storage

    config = ctx.obj["config"]

    for key in ("scheduler", "journal", "storage", "reload_after_days"):
        if not config.get(key):
            raise ValueError(f"{key} not configured")
    for key in ("metadata_fetcher_credentials",):
        if key not in config:
            raise ValueError(f"{key} not configured")

    scheduler = get_scheduler(**config["scheduler"])
    storage = get_storage(**config["storage"])

    metadata_fetcher_credentials = config["metadata_fetcher_credentials"] or {}
    reload_after_days = int(config["reload_after_days"])

    journal_cfg = config["journal"]
    journal_cfg["stop_after_objects"] = stop_after_objects or journal_cfg.get(
        "stop_after_objects"
    )

    client = get_journal_client(
        cls="kafka",
        object_types=["origin"],
        **journal_cfg,
    )
    worker = JournalClient(
        scheduler=scheduler,
        storage=storage,
        metadata_fetcher_credentials=metadata_fetcher_credentials,
        reload_after_days=reload_after_days,
    )
    nb_messages = 0
    try:
        nb_messages = client.process(worker.process_journal_objects)
        print(f"Processed {nb_messages} message(s).")
    except KeyboardInterrupt:
        ctx.exit(0)
    else:
        print("Done.")
    finally:
        client.close()



# Copyright (C) 2020-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import tempfile
from typing import List
from unittest.mock import MagicMock

from click.testing import CliRunner
from confluent_kafka import Producer
import yaml

from swh.journal.serializers import value_to_kafka
from swh.loader.metadata.cli import swh_cli_group
from swh.loader.metadata.journal_client import JournalClient


def invoke(
    args: List[str],
    storage=None,
    kafka_server=None,
    kafka_prefix=None,
    kafka_consumer_group=None,
    config=None,
):
    runner = CliRunner()

    config = dict(config or {})

    assert (
        (kafka_server is None)
        == (kafka_prefix is None)
        == (kafka_consumer_group is None)
    )
    if kafka_server:
        config["journal"] = dict(
            brokers=kafka_server,
            group_id=kafka_consumer_group,
            prefix=kafka_prefix,
            stop_on_eof=True,
        )

    with tempfile.NamedTemporaryFile("a", suffix=".yml") as config_fd:
        yaml.dump(config, config_fd)
        config_fd.seek(0)
        args = ["metadata-loader", "-C" + config_fd.name] + list(args)
        result = runner.invoke(swh_cli_group, args, catch_exceptions=False)
    return result


def test_journal_client(
    mocker,
    swh_storage,
    swh_storage_backend_config,
    swh_scheduler,
    swh_scheduler_config,
    kafka_server,
    kafka_prefix,
    kafka_consumer_group,
) -> None:
    origin_url = "http://example.org/repo.git"
    producer = Producer(
        {
            "bootstrap.servers": kafka_server,
            "client.id": "test metadata loader origin producer",
            "acks": "all",
        }
    )
    origin = {"url": origin_url}
    value = value_to_kafka(origin)
    topic = f"{kafka_prefix}.origin"
    producer.produce(topic=topic, key=b"bogus-origin", value=value)
    producer.flush()

    config = {
        "scheduler": swh_scheduler_config,
        "storage": swh_storage_backend_config,
        "metadata_fetcher_credentials": None,
        "reload_after_days": 42,
    }

    storage = MagicMock(wraps=swh_storage)
    mocker.patch("swh.storage.get_storage", return_value=storage)
    scheduler = MagicMock(wraps=swh_scheduler)
    mocker.patch("swh.scheduler.get_scheduler", return_value=scheduler)
    kwargs = dict(
        scheduler=scheduler,
        storage=storage,
        metadata_fetcher_credentials={},
        reload_after_days=42,
    )
    journal_client = JournalClient(**kwargs)  # type: ignore[arg-type]
    journal_client = MagicMock(wraps=journal_client)
    JournalClient_mock = mocker.patch(
        "swh.loader.metadata.journal_client.JournalClient", return_value=journal_client
    )
    result = invoke(
        ["journal-client"],
        kafka_server=kafka_server,
        kafka_prefix=kafka_prefix,
        kafka_consumer_group=kafka_consumer_group,
        config=config,
    )
    assert result.exit_code == 0, result.output
    assert result.output == "Processed 1 message(s).\nDone.\n"

    # Check the client was correctly initialized
    JournalClient_mock.assert_called_once_with(**kwargs)
    journal_client.process_journal_objects.assert_called_once_with(
        {"origin": [{"url": origin_url}]}
    )

    # Smoke check of the client's functionality
    scheduler.get_listed_origins.assert_called_once_with(
        url=origin_url, page_token=None
    )



# Copyright (C) 2019-2022  The Software Heritage developers
# See the AUTHORS file at the top-level directory of this distribution
# License: GNU General Public License version 3, or any later version
# See top-level LICENSE file for more information

import pkg_resources

try:
    __version__ = pkg_resources.get_distribution("swh.loader.metadata").version
except pkg_resources.DistributionNotFound:
    __version__ = "devel"


USER_AGENT_TEMPLATE = "Software Heritage Metadata Loader (%s)"
USER_AGENT = USER_AGENT_TEMPLATE % __version__
